### R code from vignette source 'opm-growth-curves.Rnw'

###################################################
### code chunk number 1: chunk-setup
###################################################
options(width = 90)
options(prompt = "R> ")
options(continue = "   ")
options(useFancyQuotes = FALSE)
library("methods")


###################################################
### code chunk number 2: load-library
###################################################
if ("package:opm" %in% search())
  detach(name = package:opm, unload = TRUE)
library("opm")


###################################################
### code chunk number 3: load-potatoes
###################################################
data(potato)
head(potato)


###################################################
### code chunk number 4: reshape-potatoes
###################################################
potato.fm <- reshape(data = potato, direction = "wide",
  v.names = "FM", drop = "DM",
  idvar = c("Genotype", "Treatment", "Replicate"),
  timevar = "Time")
potato.dm <- reshape(data = potato, direction = "wide",
  v.names = "DM", drop = "FM",
  idvar = c("Genotype", "Treatment", "Replicate"),
  timevar = "Time")


###################################################
### code chunk number 5: potato-reshaped
###################################################
head(potato.fm)


###################################################
### code chunk number 6: potatoes-to-opmx
###################################################
potato.fm <- opmx(object = potato.fm,
  position = c("Genotype", "Replicate"),
  well = "Treatment", prefix = "FM.",
  full.name = c(fm = "Growth experiment, fresh mass"))
potato.dm <- opmx(object = potato.dm,
  position = c("Genotype", "Replicate"),
  well = "Treatment", prefix = "DM.",
  full.name = c(dm = "Growth experiment, dry mass"))


###################################################
### code chunk number 7: show-potato-plate-type
###################################################
plate_type(TRUE) # shows all existing user-defined plates
listing(wells(plate = c("CUSTOM:FM", "CUSTOM:DM")))


###################################################
### code chunk number 8: register_plate
###################################################
register_plate(
  growth = c(
    A01 = "Negative Control #1", A02 = "10mM Glucose #1",
    A03 = "20mM Glucose #1", A04 = "50mM Glucose #1",
    A05 = "100mM Glucose #1", A06 = "200mM Glucose #1",
    B01 = "Negative Control #2", B02 = "10mM Glucose #2",
    B03 = "20mM Glucose #2", B04 = "50mM Glucose #2",
    B05 = "100mM Glucose #2", B06 = "200mM Glucose #2",
    C01 = "Negative Control #3", C02 = "10mM Glucose #3",
    C03 = "20mM Glucose #3", C04 = "50mM Glucose #3",
    C05 = "100mM Glucose #3", C06 = "200mM Glucose #3",
    D01 = "Negative Control #4", D02 = "10mM Glucose #4",
    D03 = "20mM Glucose #4", D04 = "50mM Glucose #4",
    D05 = "100mM Glucose #4", D06 = "200mM Glucose #4"
  ),
  growth = "Growth on Glucose"
)
listing(wells(plate = "custom:growth"))


###################################################
### code chunk number 9: register-plate-via-matrix
###################################################
# create constant part of the substrate names
growth <- c("Negative Control", "10mM Glucose", "20mM Glucose",
  "50mM Glucose", "100mM Glucose", "200mM Glucose")
# create repetitions and assign according numbers
growth <- paste(rep(growth, each = 4), rep(1:4, 4), sep = " #")
# create matrix that mirrors the plate layout
growth <- matrix(growth, nrow = 4, ncol = 6,
  dimnames = list(LETTERS[1:4], 1:6))
# register this plate type and show the result
register_plate(growth = growth, growth = "Growth on Glucose")
listing(wells(plate = "custom:growth"))


###################################################
### code chunk number 10: register-plate-via-data-frame
###################################################
register_plate(growth = NULL)
growth <- as.data.frame(growth)
register_plate(growth = growth, growth = "Growth on Glucose")
listing(wells(plate = "CUSTOM:GROWTH"))


###################################################
### code chunk number 11: input-tecan
###################################################
tecan.file <- opm_files("growth")
tecan.file <- grep(pattern = "tecan", x = tecan.file,
  ignore.case = TRUE, value = TRUE)
tecan <- read.table(tecan.file)
head(tecan)


###################################################
### code chunk number 12: opmx-rectangular
###################################################
tecan <- opmx(object = tecan, format = "rectangular",
  plate.type = "growth", position = "1A", interval = 1)
tecan


###################################################
### code chunk number 13: split-tecan
###################################################
tecan <- split(tecan)
metadata(tecan) <- data.frame(Replicate = c(1, 2, 1, 2),
  Strain = rep(c("DSM30083", "DSM18039"), each = 2),
  stringsAsFactors = FALSE)
dim(tecan)


###################################################
### code chunk number 14: look-at-perkin-elmer
###################################################
pe.file <- opm_files("growth")
pe.file <- grep(pattern = "perkin_elmer", x = pe.file,
  ignore.case = TRUE, value = TRUE)
pe <- readLines(pe.file)
cat(head(pe), sep = "\n")
cat(tail(pe), sep = "\n")


###################################################
### code chunk number 15: input-perkin-elmer
###################################################
pe <- read.table(file = pe.file, header = TRUE,
  strip.white = TRUE, sep = "\t",
  # fix the column data types and skip some columns
  colClasses = c("integer", "integer", "character", "NULL", "NULL", "numeric"),
  # re-use names but adapt last column to opmx()
  col.names = c("Plate", "Repeat", "Well", "Skip", "Skip", "T_1"),
  # skip all irregular rows after the tab-separated part
  nrows = 192L,
  # use comma as decimal separator
  dec = ",")
head(pe)


###################################################
### code chunk number 16: convert-perkin-elmer
###################################################
# copy the plate type
register_plate(myeco = wells(plate = "eco"))
# convert to OPMS
pe <- opmx(object = pe, format = "horizontal",
  full.name = c("custom:myeco" = "Ecoplate"),
  position = "Repeat", well = "Well", file = pe.file)
pe <- do_aggr(pe)


###################################################
### code chunk number 17: plot-perkin-elmer-code (eval = FALSE)
###################################################
## heat_map(pe, NULL) # no metadata available


###################################################
### code chunk number 18: plot-perkin-elmer
###################################################
heat_map(pe, NULL) # no metadata available


###################################################
### code chunk number 19: plot-perkin-elmer-split-code (eval = FALSE)
###################################################
## pe.sep <- split(pe)
## heat_map(object = pe.sep, as.labels = NULL,
##   extract.args = list(rm.num = TRUE))


###################################################
### code chunk number 20: plot-perkin-elmer-split
###################################################
pe.sep <- split(pe)
heat_map(object = pe.sep, as.labels = NULL,
  extract.args = list(rm.num = TRUE))


###################################################
### code chunk number 21: add-file-to-keys (eval = FALSE)
###################################################
## opm_opt(csv.keys = c(opm_opt("csv.keys"), "File"))


###################################################
### code chunk number 22: potato-growth-code (eval = FALSE)
###################################################
## library("gridExtra")
## plot.fm <- xy_plot(x = potato.fm, theor.max = FALSE, rcr = 1,
##   include = "Genotype", main = list(in.parens = FALSE),
##   ylab = "Fresh cell mass [mg]", neg.ctrl = FALSE)
## plot.dm <- xy_plot(x = potato.dm, theor.max = FALSE, rcr = 1,
##   include = "Genotype", main = list(in.parens = FALSE),
##   ylab = "Dry mass [mg]", neg.ctrl = FALSE)
## grid.arrange(plot.fm, plot.dm, ncol = 2)


###################################################
### code chunk number 23: potato-growth-figure
###################################################
library("gridExtra")
plot.fm <- xy_plot(x = potato.fm, theor.max = FALSE, rcr = 1,
  include = "Genotype", main = list(in.parens = FALSE),
  ylab = "Fresh cell mass [mg]", neg.ctrl = FALSE)
plot.dm <- xy_plot(x = potato.dm, theor.max = FALSE, rcr = 1,
  include = "Genotype", main = list(in.parens = FALSE),
  ylab = "Dry mass [mg]", neg.ctrl = FALSE)
grid.arrange(plot.fm, plot.dm, ncol = 2)


###################################################
### code chunk number 24: xy-plot-coli-code (eval = FALSE)
###################################################
## xy_plot(x = tecan, theor.max = FALSE, include = "Strain",
##   main = list(in.parens = FALSE), ylab = "OD at 690nm", rm.num = TRUE)


###################################################
### code chunk number 25: xy-plot-coli-figure
###################################################
xy_plot(x = tecan, theor.max = FALSE, include = "Strain",
  main = list(in.parens = FALSE), ylab = "OD at 690nm", rm.num = TRUE)


###################################################
### code chunk number 26: spline-helper-function
###################################################
# A function not normally needed by the user!
plot_param_diff <- function(...) {
  x <- list(...)
  if (is.null(names(x)))
    names(x) <- seq_along(x)
  for (param in rev(param_names())) {
    y <- lapply(lapply(x, aggregated, param, FALSE), unlist)
    n <- names(y)
    for (i in seq_along(y)[-1L])
      for (j in seq.int(1L, i - 1L)) {
        plot(x = y[[i]], y = y[[j]], pch = 19L, col = "darkgrey",
          main = sprintf("%s, %s/%s", param, n[[i]], n[[j]]))
        abline(a = line(y[[i]], y[[j]]), lty = "dashed")
      }
  }
  invisible(NULL)
}


###################################################
### code chunk number 27: agg-opm-spline
###################################################
tp.tecan <- do_aggr(object = tecan, method = "splines",
  options = set_spline_options(type = "tp.spline"))
p.tecan <- do_aggr(object = tecan, method = "splines",
  options = set_spline_options(type = "p.spline"))
sm.tecan <- do_aggr(object = tecan, method = "splines",
  options = set_spline_options(type = "smooth.spline"))


###################################################
### code chunk number 28: plot-opm-spline-code (eval = FALSE)
###################################################
## old.par <- par(mfrow = c(4, 3), mar = c(2, 2, 3, 2))
## plot_param_diff(Tp = tp.tecan, P = p.tecan, Sm = sm.tecan)
## par(old.par)


###################################################
### code chunk number 29: plot-opm-spline
###################################################
old.par <- par(mfrow = c(4, 3), mar = c(2, 2, 3, 2))
plot_param_diff(Tp = tp.tecan, P = p.tecan, Sm = sm.tecan)
par(old.par)


###################################################
### code chunk number 30: select-splines
###################################################
tecan <- tp.tecan
rm(tp.tecan, p.tecan, sm.tecan) # tidy up
potato.fm <- do_aggr(object = potato.fm, method = "splines",
  options = set_spline_options(type = "smooth.spline"))
potato.dm <- do_aggr(object = potato.dm, method = "splines",
  options = set_spline_options(type = "smooth.spline"))


###################################################
### code chunk number 31: potato-mcp
###################################################
dm.mcp <- opm_mcp(object = potato.dm, ~ J(Well, Genotype), m.type = "aov",
  linfct = c(Pairs.Well = 1), max = 7, in.parens = FALSE)
fm.mcp <- opm_mcp(object = potato.fm, ~ J(Well, Genotype), m.type = "aov",
  linfct = c(Pairs.Well = 1), max = 7, in.parens = FALSE)


###################################################
### code chunk number 32: plot-potato-mcp-code (eval = FALSE)
###################################################
## old.par <- par(mfrow = c(2, 1), mar = c(1, 15, 2, 1))
## plot(fm.mcp)
## plot(dm.mcp, main = "")
## par(old.par)


###################################################
### code chunk number 33: plot-potato-mcp
###################################################
old.par <- par(mfrow = c(2, 1), mar = c(1, 15, 2, 1))
plot(fm.mcp)
plot(dm.mcp, main = "")
par(old.par)


###################################################
### code chunk number 34: tecan-mcp
###################################################
tecan.mcp <- opm_mcp(object = tecan, ~ J(Well, Strain),
  m.type = "aov", linfct = c(Pairs.Well = 1), full = FALSE)


###################################################
### code chunk number 35: plot-tecan-mcp-code (eval = FALSE)
###################################################
## old.mar <- par(mar = c(3, 15, 3, 2))
## plot(tecan.mcp)
## par(old.mar)


###################################################
### code chunk number 36: plot-tecan-mcp
###################################################
old.mar <- par(mar = c(3, 15, 3, 2))
plot(tecan.mcp)
par(old.mar)


